#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2023-2025 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path
from typing import Any, Generator


class FoundSymbolInfo:
    section: Section
    symbol: Symbol
    offset: int

    def __init__(self, section: Section, symbol: Symbol, offset: int=0): ...

    def getAsStr(self) -> str: ...
    def getAsStrPlusOffset(self, symName: str|None=None) -> str: ...


class SymbolComparisonInfo:
    symbol: Symbol
    buildAddress: int
    buildFile: Section|None
    expectedAddress: int
    expectedFile: Section|None

    def __init__(self, symbol: Symbol, buildAddress: int, buildFile: Section|None, expectedAddress: int, expectedFile: Section|None): ...

    def diff(self) -> int|None: ...

class MapsComparisonInfo:
    badFiles: set[Section] = set()
    missingFiles: set[Section] = set()
    comparedList: list[SymbolComparisonInfo] = []

    def __init__(self): ...


class ProgressStats:
    undecompedSize: int
    decompedSize: int

    def __init__(self): ...

    @property
    def total(self) -> int: ...

    def undecompedPercentage(self) -> float:
        ...

    def decompedPercentage(self) -> float:
        ...

    def undecompedPercentageTotal(self, totalStats: ProgressStats) -> float:
        ...

    def decompedPercentageTotal(self, totalStats: ProgressStats) -> float:
        ...

    def getAsFrogressEntry(self, name: str) -> dict[str, int]: ...

    @staticmethod
    def getHeaderAsStr(categoryColumnSize: int=28) -> str:
        ...

    @staticmethod
    def printHeader(categoryColumnSize: int=28) -> None: ...

    def getEntryAsStr(self, category: str, totalStats: ProgressStats, categoryColumnSize: int=28) -> str:
        ...

    def print(self, category: str, totalStats: ProgressStats, categoryColumnSize: int=28) -> None: ...


class Symbol:
    name: str
    vram: int
    size: int # in bytes
    vrom: int|None
    align: int|None
    nonmatchingSymExists: bool

    def __init__(self, name: str, vram: int, size: int=0, vrom: int|None=None, align: int|None=None, nonmatchingSymExists: bool=False): ...

    def getVramStr(self) -> str: ...
    def getSizeStr(self) -> str: ...
    def getVromStr(self) -> str: ...
    def getAlignStr(self) -> str: ...

    def serializeName(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVram(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeSize(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVrom(self, humanReadable: bool=True) -> str|int|None: ...

    @staticmethod
    def toCsvHeader() -> str: ...
    def toCsv(self) -> str: ...

    @staticmethod
    def printCsvHeader(): ...
    def printAsCsv(self): ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def __eq__(self, other) -> bool: ...
    def __hash__(self): ...
    def __repr__(self): ...


class Section:
    filepath: Path
    vram: int
    size: int # in bytes
    sectionType: str
    vrom: int|None
    align: int|None
    isFill: bool
    # _symbols: list[Symbol]

    def __init__(self, filepath: Path, vram: int, size: int, section_type: str, vrom: int|None=None, align: int|None=None, isFill: bool=False) -> None: ...

    @property
    def isNoloadSection(self) -> bool: ...

    def serializeVram(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeSize(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVrom(self, humanReadable: bool=True) -> str|int|None: ...

    #! @deprecated
    def getName(self) -> Path: ...

    def findSymbolByName(self, symName: str) -> Symbol|None: ...
    #! @deprecated: Use either `findSymbolByVram` or `findSymbolByVrom` instead.
    def findSymbolByVramOrVrom(self, address: int) -> tuple[Symbol, int]|None: ...
    def findSymbolByVram(self, address: int) -> tuple[Symbol, int]|None: ...
    def findSymbolByVrom(self, address: int) -> tuple[Symbol, int]|None: ...

    @staticmethod
    def toCsvHeader(printVram: bool=True) -> str: ...
    def toCsv(self, printVram: bool=True) -> str: ...

    @staticmethod
    def printCsvHeader(printVram: bool=True): ...
    def printAsCsv(self, printVram: bool=True): ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def copySymbolList(self) -> list[Symbol]:
        """Returns a copy (not a reference) of the internal symbol list"""

    def setSymbolList(self, newList: list[Symbol]) -> None:
        """Replaces the internal symbol list with a copy of `newList`"""

    def appendSymbol(self, sym: Symbol) -> None:
        """Appends a copy of `sym` into the internal symbol list"""

    def __iter__(self) -> Generator[Symbol, None, None]: ...

    def __getitem__(self, index) -> Symbol: ...
    def __setitem__(self, index, sym: Symbol): ...
    def __len__(self) -> int: ...
    def __eq__(self, other) -> bool: ...
    def __hash__(self): ...


class Segment:
    name: str
    vram: int
    size: int
    vrom: int
    align: int|None
    # _filesList: list[Section] = dataclasses.field(default_factory=list)

    def __init__(self, name: str, vram: int, size: int, vrom: int, align: int|None=None): ...

    def serializeVram(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeSize(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVrom(self, humanReadable: bool=True) -> str|int|None: ...

    def filterBySectionType(self, sectionType: str) -> Segment: ...

    def getEverySectionExceptSectionType(self, sectionType: str) -> Segment: ...
    #! @deprecated: Use either `getEverySectionExceptSectionType` instead.
    def getEveryFileExceptSectionType(self, sectionType: str) -> Segment: ...

    def findSymbolByName(self, symName: str) -> FoundSymbolInfo|None: ...
    #! @deprecated: Use either `findSymbolByVram` or `findSymbolByVrom` instead.
    def findSymbolByVramOrVrom(self, address: int) -> FoundSymbolInfo|None: ...
    def findSymbolByVram(self, address: int) -> tuple[FoundSymbolInfo|None, list[Section]]: ...
    def findSymbolByVrom(self, address: int) -> tuple[FoundSymbolInfo|None, list[Section]]: ...

    def mixFolders(self) -> Segment: ...

    def printAsCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True): ...
    def printSymbolsCsv(self): ...

    def toCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True) -> str: ...
    def toCsvSymbols(self) -> str: ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def copySectionList(self) -> list[Section]:
        """Returns a copy (not a reference) of the internal section list"""

    def setSectionList(self, newList: list[Section]) -> None:
        """Replaces the internal section list with a copy of `newList`"""

    def appendSection(self, section: Section) -> None:
        """Appends a copy of `section` into the internal section list"""

    #! @deprecated: Use either `copySectionList` instead.
    def copyFileList(self) -> list[Section]:
        """Returns a copy (not a reference) of the internal section list"""

    #! @deprecated: Use either `setSectionList` instead.
    def setFileList(self, newList: list[Section]) -> None:
        """Replaces the internal section list with a copy of `newList`"""

    #! @deprecated: Use either `appendSection` instead.
    def appendFile(self, section: Section) -> None:
        """Appends a copy of `section` into the internal section list"""

    def __iter__(self) -> Generator[Section, None, None]: ...
    def __getitem__(self, index) -> Section: ...
    def __setitem__(self, index, section: Section): ...
    def __len__(self) -> int: ...
    def __eq__(self, other) -> bool: ...
    def __hash__(self): ...


class MapFile:
    #! @deprecated: Use either `newFromMapFile` or `newFromMapStr` instead.
    def __init__(self) -> None: ...

    @staticmethod
    def newFromMapFile(mapPath: Path) -> MapFile: ...
    @staticmethod
    def newFromMapStr(mapContents: str) -> MapFile: ...
    @staticmethod
    def newFromGnuMapStr(mapContents: str) -> MapFile: ...
    @staticmethod
    def newFromLldMapStr(mapContents: str) -> MapFile: ...

    #! @deprecated: Use either `newFromMapFile` instead.
    def readMapFile(self, mapPath: Path) -> None: ...
    #! @deprecated: Use either `newFromMapStr` instead.
    def parseMapContents(self, mapContents: str) -> None: ...
    #! @deprecated: Use either `newFromGnuMapStr` instead.
    def parseMapContentsGNU(self, mapContents: str) -> None: ...
    #! @deprecated: Use either `newFromLldMapStr` instead.
    def parseMapContentsLLD(self, mapContents: str) -> None: ...

    def filterBySectionType(self, sectionType: str) -> MapFile: ...

    def getEverySectionExceptSectionType(self, sectionType: str) -> MapFile: ...
    #! @deprecated: Use either `getEverySectionExceptSectionType` instead.
    def getEveryFileExceptSectionType(self, sectionType: str) -> MapFile: ...

    def findSymbolByName(self, symName: str) -> FoundSymbolInfo|None: ...
    #! @deprecated: Use either `findSymbolByVram` or `findSymbolByVrom` instead.
    def findSymbolByVramOrVrom(self, address: int) -> FoundSymbolInfo|None: ...
    def findSymbolByVram(self, address: int) -> tuple[FoundSymbolInfo|None, list[Section]]: ...
    def findSymbolByVrom(self, address: int) -> tuple[FoundSymbolInfo|None, list[Section]]: ...

    def findLowestDifferingSymbol(self, otherMapFile: MapFile) -> tuple[Symbol, Section, Symbol|None]|None: ...

    def mixFolders(self) -> MapFile: ...

    def getProgress(self, asmPath: Path, nonmatchings: Path, aliases: dict[str, str]=dict(), pathIndex: int=2) -> tuple[ProgressStats, dict[str, ProgressStats]]: ...

    # Useful for finding bss reorders
    def compareFilesAndSymbols(self, otherMapFile: MapFile, *, checkOtherOnSelf: bool=True) -> MapsComparisonInfo: ...

    def toCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True) -> str: ...
    def toCsvSymbols(self) -> str: ...

    def printAsCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True): ...
    def printSymbolsCsv(self): ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def copySegmentList(self) -> list[Segment]:
        """Returns a copy (not a reference) of the internal segment list"""

    def setSegmentList(self, newList: list[Segment]) -> None:
        """Replaces the internal segment list with a copy of `newList`"""

    def appendSegment(self, segment: Segment) -> None:
        """Appends a copy of `segment` into the internal segment list"""

    def __iter__(self) -> Generator[Segment, None, None]: ...
    def __getitem__(self, index) -> Segment: ...
    def __setitem__(self, index, segment: Segment): ...
    def __len__(self) -> int: ...
